#ifndef _GSH_OPENGL_H_
#define _GSH_OPENGL_H_

#include "GSHandler.h"
#include "win32/Window.h"
#include <gl/gl.h>

class CGSH_OpenGL : public CGSHandler
{
public:
									CGSH_OpenGL(Framework::CWindow*);
									~CGSH_OpenGL();

	static void						CreateGSHandler(Framework::CWindow*);

	void							WriteRegister(uint8, uint64);
	void							ProcessImageTransfer(uint32, uint32);
	void							SetCrt(int, int, int);

	void							Flip();

private:
	enum MAXCACHE
	{
		MAXCACHE = 5,
	};

	struct VERTEX
	{
		uint64						nPosition;
		uint64						nRGBAQ;
		uint64						nUV;
		uint64						nST;
	};

	class CTexture
	{
	public:
									CTexture();
									~CTexture();
		void						Invalidate();
		void						InvalidateFromMemorySpace(uint32, uint32);
		bool						IsValid();

		uint32						m_nStart;
		uint32						m_nSize;
		unsigned int				m_nTexture;
	};

	static CGSHandler*				GSHandlerFactory(void*);

	void							InitializeRC();
	void							LinearZOrtho(double, double, double, double);
	void							CreateViewport(int, int);
	void							UpdateViewport();
	unsigned int					LoadTexture(GSTEX0*, GSTEX1*);

	double							GetZ(double);

	void							Prim_Point();
	void							Prim_Line();
	void							Prim_Triangle();
	void							Prim_Sprite();

	void							SetRenderingContext(unsigned int);
	void							SetupTestFunctions(uint64);
	void							SetupBlendingFunction(uint64);
	void							SetupTexture(uint64, uint64);

	//Context variables (put this in a struct or something?)
	double							m_nPrimOfsX;
	double							m_nPrimOfsY;
	uint32							m_nTexWidth;
	uint32							m_nTexHeight;
	unsigned int					m_nTexHandle;

	int								m_nWidth;
	int								m_nHeight;

	unsigned int					TexCache_Search(uint32);
	void							TexCache_Insert(uint32, uint32, unsigned int);
	void							TexCache_InvalidateTextures(uint32, uint32);

	CTexture						m_TexCache[MAXCACHE];
	unsigned int					m_nTexCacheIndex;

	VERTEX							m_VtxBuffer[3];
	int								m_nVtxCount;

	GSPRMODE						m_PrimitiveMode;
	unsigned int					m_nPrimitiveType;

	Framework::CWindow*				m_pOutputWnd;

	HGLRC							m_hRC;
	HDC								m_hDC;
	static PIXELFORMATDESCRIPTOR	m_PFD;

};

#endif