#ifndef _GS_H_
#define _GS_H_

#include "Types.h"
#include "Stream.h"

#pragma pack(push, 1)

enum GS_REGS
{
	GS_REG_PRIM			= 0x00,
	GS_REG_RGBAQ		= 0x01,
	GS_REG_ST			= 0x02,
	GS_REG_UV			= 0x03,
	GS_REG_XYZ2			= 0x05,
	GS_REG_TEX0_1		= 0x06,
	GS_REG_TEX0_2		= 0x07,
	GS_REG_TEX1_1		= 0x14,
	GS_REG_TEX1_2		= 0x15,
	GS_REG_XYOFFSET_1	= 0x18,
	GS_REG_XYOFFSET_2	= 0x19,
	GS_REG_PRMODECONT	= 0x1A,
	GS_REG_PRMODE		= 0x1B,
	GS_REG_TEXFLUSH		= 0x3F,
	GS_REG_SCISSOR_1	= 0x40,
	GS_REG_ALPHA_1		= 0x42,
	GS_REG_ALPHA_2		= 0x43,
	GS_REG_TEST_1		= 0x47,
	GS_REG_TEST_2		= 0x48,
	GS_REG_FRAME_1		= 0x4C,
	GS_REG_FRAME_2		= 0x4D,
	GS_REG_BITBLTBUF	= 0x50,
	GS_REG_TRXPOS		= 0x51,
	GS_REG_TRXREG		= 0x52,
	GS_REG_TRXDIR		= 0x53,
};

struct GSTEX0
{
	unsigned int	nBufPtr			: 14;
	unsigned int	nBufWidth		: 6;
	unsigned int	nPsm			: 6;
	unsigned int	nWidth			: 4;
	unsigned int	nPad0			: 2;
	unsigned int	nPad1			: 2;
	unsigned int	nColorComp		: 1;
	unsigned int	nFunction		: 2;
	uint32			GetBufPtr()		{ return nBufPtr * 256; }
	uint32			GetBufWidth()	{ return nBufWidth * 64; }
	uint32			GetWidth()		{ return (1 << nWidth); }
	uint32			GetHeight()		{ return (1 << (nPad0 | (nPad1 << 2))); }
};

struct GSTEX1
{
	unsigned int	nLODMethod		: 1;
	unsigned int	nReserved0		: 1;
	unsigned int	nMaxMip			: 3;
	unsigned int	nMagFilter		: 1;
	unsigned int	nMinFilter		: 3;
	unsigned int	nMipBaseAddr	: 1;
	unsigned int	nReserved1		: 9;
	unsigned int	nLODL			: 2;
	unsigned int	nReserved2		: 11;
	unsigned int	nLODK			: 7;
	unsigned int	nReserved3		: 25;
};

struct GSRGBAQ
{
	uint8			nR;
	uint8			nG;
	uint8			nB;
	uint8			nA;
	float			nQ;
};

struct GSDISPLAY
{
	unsigned int	nX			: 12;
	unsigned int	nY			: 11;
	unsigned int	nMagX		: 4;
	unsigned int	nMagY		: 5;
	unsigned int	nW			: 12;
	unsigned int	nH			: 12;
};

struct GSDISPFB
{
	unsigned int	nBufPtr			: 9;
	unsigned int	nBufWidth		: 6;
	unsigned int	nPSM			: 5;
	unsigned int	nReserved0		: 12;
	unsigned int	nX				: 11;
	unsigned int	nY				: 11;
	unsigned int	nReserved1		: 10;
	uint32			GetBufPtr()		{ return nBufPtr * 8192; };
	uint32			GetBufWidth()	{ return nBufWidth * 64; };
};

struct GSPRIM
{
	unsigned int	nType			: 3;
	unsigned int	nShading		: 1;
	unsigned int	nTexture		: 1;
	unsigned int	nFog			: 1;
	unsigned int	nAlpha			: 1;
	unsigned int 	nAntiAliasing	: 1;
	unsigned int	nUseUV			: 1;
	unsigned int	nContext		: 1;
	unsigned int	nUseFloat		: 1;
};

struct GSPRMODE
{
	unsigned int	nReserved0		: 3;
	unsigned int	nShading		: 1;
	unsigned int	nTexture		: 1;
	unsigned int	nFog			: 1;
	unsigned int	nAlpha			: 1;
	unsigned int 	nAntiAliasing	: 1;
	unsigned int	nUseUV			: 1;
	unsigned int	nContext		: 1;
	unsigned int	nUseFloat		: 1;
};

struct GSBITBLTBUF
{
	unsigned int	nSrcPtr			: 16;
	unsigned int	nSrcWidth		: 8;
	unsigned int	nSrcPsm			: 8;
	unsigned int	nDstPtr			: 16;
	unsigned int	nDstWidth		: 8;
	unsigned int	nDstPsm			: 8;
	uint32			GetSrcPtr()		{ return nSrcPtr * 256; }
	uint32			GetSrcWidth()	{ return nSrcWidth * 64; }
	uint32			GetDstPtr()		{ return nDstPtr * 256; }
	uint32			GetDstWidth()	{ return nDstWidth * 64; }
};

struct GSALPHA
{
	unsigned int	nA			: 2;
	unsigned int	nB			: 2;
	unsigned int	nC			: 2;
	unsigned int	nD			: 2;
	unsigned int	nReserved0	: 24;
	unsigned int	nFix		: 8;
	unsigned int	nReserved1	: 24;
};

struct GSTEST
{
	unsigned int	nAlphaEnabled		: 1;
	unsigned int	nAlphaMethod		: 3;
	unsigned int	nAlphaRef			: 8;
	unsigned int	nAlphaFail			: 2;
	unsigned int	nDestAlphaEnabled	: 1;
	unsigned int	nDestAlphaMode		: 1;
	unsigned int	nDepthEnabled		: 1;
	unsigned int	nDepthMethod		: 2;
	unsigned int	nReserved0			: 13;
	uint32			nReserved1;
};

struct GSFRAME
{
	unsigned int	nPtr			: 16;
	unsigned int	nWidth			: 8;
	unsigned int	nPsm			: 8;
	unsigned int	nMask			: 32;
	uint32			GetBasePtr()	{ return nPtr * 8192; }
	uint32			GetWidth()		{ return nWidth * 64; }
};

#pragma pack(pop)

#define DECODE_DISPLAY(v, d)					\
	(d) = *(GSDISPLAY*)&(v);

#define DECODE_DISPFB(v, d)						\
	(d) = *(GSDISPFB*)&(v);

#define DECODE_PRIM(v, pr)						\
	(pr) = *(GSPRIM*)&(v);

#define DECODE_PRMODE(v, prm)					\
	(prm) = *(GSPRMODE*)&(v);

#define DECODE_TEX0(v, tex)						\
	(tex) = *(GSTEX0*)&(v);

#define DECODE_TEX1(v, tex)						\
	(tex) = *(GSTEX1*)&(v);

#define DECODE_RGBAQ(v, rgbaq)					\
	(rgbaq) = *(GSRGBAQ*)&(v);

#define DECODE_UV(p, u, v)						\
	(u)  = (double)((p >>  4) & 0xFFF);			\
	(u) += (double)((p >>  0) & 0xF) / 16.0f;	\
	(v)  = (double)((p >> 20) & 0xFFF);			\
	(v) += (double)((p >> 16) & 0xF) / 16.0f;

#define DECODE_ST(p, s, t)						\
	(s)  = *(float*)((uint32*)&(p) + 0);		\
	(t)  = *(float*)((uint32*)&(p) + 1);

#define DECODE_XYZ2(v, x, y, z)					\
	(x)  = (double)((v >>  4) & 0xFFF);			\
	(x) += (double)((v >>  0) & 0xF) / 16.0f;	\
	(y)  = (double)((v >> 20) & 0xFFF);			\
	(y) += (double)((v >> 16) & 0xF) / 16.0f;	\
	(z)  = (double)((v >> 32) & 0xFFFFFFFF);

#define DECODE_XYOFFSET(v, x, y)				\
	(x)  = (double)((v >>  4) & 0xFFF);			\
	(x) += (double)((v >>  0) & 0xF) / 16.0f;	\
	(y)  = (double)((v >> 36) & 0xFFF);			\
	(y) += (double)((v >> 32) & 0xF) / 16.0f;	

#define DECODE_ALPHA(v, alpha)					\
	(alpha) = *(GSALPHA*)&(v);

#define DECODE_TEST(v, tst)						\
	(tst) = *(GSTEST*)&(v);

#define DECODE_FRAME(v, fr)						\
	(fr) = *(GSFRAME*)&(v);

#define DECODE_BITBLTBUF(v, buf)				\
	(buf) = *(GSBITBLTBUF*)&(v);

#define DECODE_TRXREG(v, w, h)					\
	(w) = (uint32)(v >>  0) & 0xFFFFFFFF;		\
	(h) = (uint32)(v >> 32) & 0xFFFFFFFF;

#define DECODE_TRXPOS(v, sx, sy, dx, dy, d)		\
	(sx) = (uint32)(v >>  0) & 0xFFFF;			\
	(sy) = (uint32)(v >> 16) & 0xFFFF;			\
	(dx) = (uint32)(v >> 32) & 0xFFFF;			\
	(dy) = (uint32)(v >> 48) & 0xFF;			\
	(d)  = (uint32)(v >> 59);

class CGSHandler
{
public:
						CGSHandler();
	virtual				~CGSHandler();

	void				SaveState(Framework::CStream*);
	void				LoadState(Framework::CStream*);

	virtual void		WritePrivRegister(uint32, uint32);
	virtual uint32		ReadPrivRegister(uint32);
	void				DisassembleWrite(uint8, uint64);
	void				SetVBlank();
	void				ResetVBlank();

	virtual void		WriteRegister(uint8, uint64);
	void				FeedImageData(void*, uint32);

	virtual void		UpdateViewport()						= 0;
	virtual void		ProcessImageTransfer(uint32, uint32)	= 0;
	virtual void		SetCrt(int, int, int)					= 0;
	virtual void		Flip()									= 0;

protected:
	enum RAMSIZE
	{
		RAMSIZE = 0x00400000,
	};

	struct TRXCONTEXT
	{
		uint32			nSize;
		uint32			nCount;
		uint32			nPtr;
		uint32			nSrcPitch;
		uint32			nDstPitch;
	};

	uint64				m_nPMODE;			//0x12000000
	uint64				m_nDISPLAY1;		//0x12000080
	uint64				m_nDISPFB2;			//0x12000090
	uint64				m_nDISPLAY2;		//0x120000A0
	uint64				m_nCSR;				//0x12001000

	TRXCONTEXT			m_TrxCtx;
	uint64				m_nReg[0x80];

	uint8*				m_pRAM;
};

#endif
