#ifndef _IOP_FILEIO_H_
#define _IOP_FILEIO_H_

#include <stdio.h>
#include "IOP_Module.h"
#include "List.h"

namespace IOP
{
	class CFileIO : public CModule
	{
	public:
									CFileIO();
									~CFileIO();
		virtual void				Invoke(uint32, void*, uint32, void*, uint32);

		enum MODULE_ID
		{
			MODULE_ID	= 0x80000001
		};

	private:
		class CFile
		{
		public:
			virtual					~CFile() {}
			virtual uint32			Read(uint32, void*) = 0;
			virtual uint32			Write(uint32, void*) = 0;
			virtual uint32			Seek(uint32, uint32) = 0;;
		};

		class CStdioFile : public CFile
		{
		public:
									CStdioFile(FILE*);
			virtual					~CStdioFile();
			virtual uint32			Read(uint32, void*);
			virtual uint32			Write(uint32, void*);
			virtual uint32			Seek(uint32, uint32);

		private:
			FILE*					m_pStream;
		};

		class CDevice
		{
		public:
									CDevice(const char*);
			virtual					~CDevice() {}
			virtual CFile*			OpenFile(uint32, const char*) = 0;
			virtual const char*		GetName();

		private:
			const char*				m_sName;
		};

		class CDirectoryDevice : public CDevice
		{
		public:
									CDirectoryDevice(const char*, const char*);
			virtual CFile*			OpenFile(uint32, const char*);
		private:
			const char*				m_sBasePath;
		};

		enum OPEN_FLAGS
		{
			O_RDONLY	= 0x00000001,
			O_WRONLY	= 0x00000002,
			O_RDWR		= 0x00000003,
			O_CREAT		= 0x00000200,
		};

		bool						SplitPath(const char*, char*, char*);
		CDevice*					FindDevice(const char*);
		uint32						RegisterFile(CFile*);

		uint32						Open(uint32, const char*);
		uint32						Close(uint32);
		uint32						Read(uint32, uint32, void*);
		uint32						Write(uint32, uint32, void*);
		uint32						Seek(uint32, uint32, uint32);

		uint32						m_nFileID;

		Framework::CList<CFile>		m_File;
		Framework::CList<CDevice>	m_Device;
	};
}

#endif
