#include "MA_EE.h"
#include "MIPS.h"
#include "PS2OS.h"

CMA_EE g_MAEE;

CMA_EE::CMA_EE() :
CMA_MIPSIV(MIPS_REGSIZE_64)
{
	m_pOpGeneral[0x1E] = LQ;
	m_pOpGeneral[0x1F] = SQ;

	//OS special instructions
	m_pOpSpecial[0x1D] = REEXCPT;

	m_pOpSpecial2[0x00] = MADD;
	m_pOpSpecial2[0x08] = MMI0;
	m_pOpSpecial2[0x09] = MMI2;
	m_pOpSpecial2[0x10] = MFHI1;
	m_pOpSpecial2[0x12] = MFLO1;
	m_pOpSpecial2[0x18] = MULT1;
	m_pOpSpecial2[0x19] = MULTU1;
	m_pOpSpecial2[0x1A] = DIV1;
	m_pOpSpecial2[0x1B] = DIVU1;
	m_pOpSpecial2[0x29] = MMI3;

	SetupReflectionTables();
}

CMA_EE::~CMA_EE()
{

}

//////////////////////////////////////////////////
//General Opcodes
//////////////////////////////////////////////////

//1E
void CMA_EE::LQ()
{
	ComputeMemAccessAddr();

	//Load the word
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRT]);	
}

//1F
void CMA_EE::SQ()
{
	ComputeMemAccessAddr();

	//Write the words
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 3, false);
}

//////////////////////////////////////////////////
//Special Opcodes
//////////////////////////////////////////////////

void CMA_EE::REEXCPT()
{
	m_pB->Call(CPS2OS::ExceptionReentry, 0, false);
}

//////////////////////////////////////////////////
//MMI Opcodes
//////////////////////////////////////////////////

//00
void CMA_EE::MADD()
{
	//Multiply
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->MultS();

	//Add LO
	m_pB->PushAddr(&m_pCtx->m_State.nLO[0]);
	m_pB->AddC();

	//Save LO/RD
	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nLO[1]);

	if(m_nRD != 0)
	{
		m_pB->PushTop();
		SignExtendTop32(m_nRD);
		m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
	}

	m_pB->PullAddr(&m_pCtx->m_State.nLO[0]);

	//Add HI
	m_pB->PushAddr(&m_pCtx->m_State.nHI[0]);
	m_pB->Add();
	m_pB->Add();

	//Save HI
	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nHI[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nHI[0]);
}

//08
void CMA_EE::MMI0()
{
	m_pOpMmi0[(m_nOpcode >> 6) & 0x1F]();
}

//09
void CMA_EE::MMI2()
{
	m_pOpMmi2[(m_nOpcode >> 6) & 0x1F]();
}

//10
void CMA_EE::MFHI1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nHI1[0]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nHI1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
}

//12
void CMA_EE::MFLO1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nLO1[0]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nLO1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
}

//18
void CMA_EE::MULT1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->MultS();

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[0]);

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[0]);

	if(m_nRD != 0)
	{
		m_pB->PushAddr(&m_pCtx->m_State.nLO1[0]);
		SignExtendTop32(m_nRD);
		m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
	}
}

//19
void CMA_EE::MULTU1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->Mult();

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[0]);

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[0]);

	if(m_nRD != 0)
	{
		m_pB->PushAddr(&m_pCtx->m_State.nLO1[0]);
		SignExtendTop32(m_nRD);
		m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
	}
}

//1A
void CMA_EE::DIV1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->DivS();

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[0]);

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[0]);
}

//1B
void CMA_EE::DIVU1()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->Div();

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nLO1[0]);

	m_pB->SeX32();
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[1]);
	m_pB->PullAddr(&m_pCtx->m_State.nHI1[0]);
}

//29
void CMA_EE::MMI3()
{
	m_pOpMmi3[(m_nOpcode >> 6) & 0x1F]();
}

//////////////////////////////////////////////////
//MMI0 Opcodes
//////////////////////////////////////////////////

//01
void CMA_EE::PSUBW()
{
	//A
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);

	//B
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);

	m_pB->VU_SubW128();

	//Result
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
}

//09
void CMA_EE::PSUBB()
{
	//A
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);

	//B
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);

	m_pB->VU_SubB128();

	//Result
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
}

//////////////////////////////////////////////////
//MMI2 Opcodes
//////////////////////////////////////////////////

//0E
void CMA_EE::PCPYLD()
{
	//A0
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);

	//B0
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
}

//12
void CMA_EE::PAND()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->And();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->And();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->And();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);
	m_pB->And();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
}

//13
void CMA_EE::PXOR()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->Xor();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->Xor();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->Xor();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);
	m_pB->Xor();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
}

//////////////////////////////////////////////////
//MMI3 Opcodes
//////////////////////////////////////////////////

//0E
void CMA_EE::PCPYUD()
{
	//A
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);
	
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	//B
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);
	
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);	
}

//13
void CMA_EE::PNOR()
{
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->Or();
	m_pB->Not();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR1[m_nRT]);
	m_pB->Or();
	m_pB->Not();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);
	
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->Or();
	m_pB->Not();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);

	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRS]);
	m_pB->PushAddr(&m_pCtx->m_State.nGPR3[m_nRT]);
	m_pB->Or();
	m_pB->Not();
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
}

//1B
void CMA_EE::PCPYH()
{
	//A0
	m_pB->PushAddr(&m_pCtx->m_State.nGPR0[m_nRT]);
	m_pB->AndImm(0xFFFF);
	m_pB->PushTop();
	m_pB->SllImm(16);
	m_pB->Or();
	m_pB->PushTop();

	m_pB->PullAddr(&m_pCtx->m_State.nGPR0[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR1[m_nRD]);

	//A1
	m_pB->PushAddr(&m_pCtx->m_State.nGPR2[m_nRT]);
	m_pB->AndImm(0xFFFF);
	m_pB->PushTop();
	m_pB->SllImm(16);
	m_pB->Or();
	m_pB->PushTop();

	m_pB->PullAddr(&m_pCtx->m_State.nGPR2[m_nRD]);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR3[m_nRD]);
}

//////////////////////////////////////////////////
//Opcode Tables
//////////////////////////////////////////////////

void (*CMA_EE::m_pOpMmi0[0x20])() = 
{
	//0x00
	Illegal,		PSUBW,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		PSUBB,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CMA_EE::m_pOpMmi2[0x20])() = 
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		PCPYLD,			Illegal,
	//0x10
	Illegal,		Illegal,		PAND,			PXOR,			Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CMA_EE::m_pOpMmi3[0x20])() = 
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		PCPYUD,			Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		PNOR,			Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		PCPYH,			Illegal,		Illegal,		Illegal,		Illegal,
};
