#include "PS2OS.h"
#include "PS2VM.h"
#include "StdStream.h"
#include "PtrMacro.h"
#include "Utils.h"
#include "DMAC.h"
#include "INTC.h"
#include "SIF.h"
#include "COP_SCU.h"

#define CONFIGPATH ".\\config\\"

using namespace Framework;

CSlot<int>							CPS2OS::m_OnExecutableChange;
CSlot<int>							CPS2OS::m_OnExecutableUnloading;

bool								CPS2OS::m_nInitialized		= false;
CELF*								CPS2OS::m_pELF				= NULL;

CList<CPS2OS::SEMAPHORE>			CPS2OS::m_Semaphore;
uint32								CPS2OS::m_nNextSemaphoreID;

CList<CPS2OS::DMACHANDLER>			CPS2OS::m_DmacHandler;
uint32								CPS2OS::m_nNextDmacHandlerID;

CList<CPS2OS::CExceptionHandler>	CPS2OS::m_ExceptionHandler;

uint32								CPS2OS::m_nStackBase	= 0;
uint32								CPS2OS::m_nHeapBase		= 0;

CStr								CPS2OS::m_sExecutablePath;
CMIPS*								CPS2OS::m_pCtx				= NULL;

CPS2OS::CPS2OS()
{
	if(m_nInitialized)
	{
		assert(0);
	}
	Initialize();
}

CPS2OS::~CPS2OS()
{
	if(!m_nInitialized)
	{
		assert(0);
	}
	Release();
}

void CPS2OS::Initialize()
{
	m_pELF = NULL;
	m_pCtx = &CPS2VM::m_EE;

	m_nNextSemaphoreID		= 0;
	m_nNextDmacHandlerID	= 0;

	m_nInitialized = true;
}

void CPS2OS::Release()
{
	while(m_DmacHandler.Count() != 0)
	{
		delete m_DmacHandler.Pull();
	}
	while(m_Semaphore.Count() != 0)
	{
		delete m_Semaphore.Pull();
	}
	while(m_ExceptionHandler.Count() != 0)
	{
		delete m_ExceptionHandler.Pull();
	}

	UnloadExecutable();
	
	m_nInitialized = false;
}

bool CPS2OS::IsInitialized()
{
	return m_nInitialized;
}

void CPS2OS::BootELF(const char* sPath)
{
	CStdStream* pStream;
	CELF* pELF;

	pStream = new CStdStream(fopen(sPath, "rb"));

	try
	{
		pELF = new CELF(pStream);
		delete pStream;
	}
	catch(const char* sError)
	{
		delete pStream;
		throw sError;
	}

	//Check for MIPS CPU
	if(pELF->m_Header.nCPU != 8)
	{
		DELETEPTR(pELF);
		throw "Invalid target CPU. Must be MIPS.";
	}

	if(pELF->m_Header.nType != 2)
	{
		DELETEPTR(pELF);
		throw "Not an executable ELF file.";
	}
	
	CPS2VM::Pause();

	UnloadExecutable();

	m_pELF = pELF;

	m_sExecutablePath = sPath;

	LoadExecutable();

	//Just a test
//	pStream = new CStdStream(fopen("./vfs/host/sjpcm.irx", "rb"));
//	pStream = new CStdStream(fopen("./vfs/host/padman.irx", "rb"));
//	pELF = new CELF(pStream);
//	memcpy(CPS2VM::m_pRAM + 0x01000000, pELF->m_pData, pELF->m_nLenght);
//	delete pELF;

	m_OnExecutableChange.Notify(0);

	printf("PS2OS: Loaded '%s' executable file.\r\n", sPath);
}

CELF* CPS2OS::GetELF()
{
	return m_pELF;
}

void CPS2OS::LoadExecutable()
{
	ELFPROGRAMHEADER* p;
//	ELFSECTIONHEADER* s;
//	const void* pData;
//	unsigned int i;
	uint32 nStart;

	CPS2VM::m_EE.InvalidateCache();

	/*
	nStart = 0xFFFFFFFF;
	for(i = 1; i < m_pELF->m_Header.nSectHeaderCount; i++)
	{
		s		= m_pELF->GetSection(i);
		pData	= m_pELF->GetSectionData(i);
		
		if(s->nStart > 0x02000000) continue;
		if(s->nType != 0x01) continue;

		if(s->nFlags & 0x02)
		{
			memcpy(CPS2VM::m_pRAM + s->nStart, pData, s->nSize);
		}

		if(s->nStart < nStart)
		{
			nStart = s->nStart;
		}
	}
	*/

	p = m_pELF->GetProgram(0);
	if(p != NULL)
	{
		nStart = p->nVAddress;

		memcpy(CPS2VM::m_pRAM + nStart, (uint8*)m_pELF->m_pData + p->nOffset, p->nFileSize);
	}

	//Load the comments maybe?
	LoadExecutableConfig();

	//InsertFunctionSymbols();

	CPS2VM::m_EE.m_State.nPC = m_pELF->m_Header.nEntryPoint;
	
	//Install hooks
	CPS2VM::m_EE.m_pSysCallHandler = SysCallHandler;
	*(uint32*)&CPS2VM::m_pBIOS[0x00000004] = 0x0000001D;

	CPS2VM::m_EE.m_pAnalysis->Clear();
	CPS2VM::m_EE.m_pAnalysis->Analyse(nStart, (nStart + m_pELF->m_nLenght) & ~0x3);

	CPS2VM::m_OnMachineStateChange.Notify(0);
}

void CPS2OS::UnloadExecutable()
{
	m_OnExecutableUnloading.Notify(0);

	if(m_pELF == NULL) return;

	DELETEPTR(m_pELF);

	SaveExecutableConfig();

	CPS2VM::m_EE.m_Comments.RemoveTags();
	CPS2VM::m_EE.m_Functions.RemoveTags();
}

void CPS2OS::LoadExecutableConfig()
{

#ifdef DEBUGGER_INCLUDED

	char sPath[256];
	const char* sFilename;

	sFilename = Utils::GetFilenameFromPath(m_sExecutablePath);

	strcpy(sPath, CONFIGPATH);
	strcat(sPath, sFilename);
	strcat(sPath, ".functions");

	CPS2VM::m_EE.m_Functions.Unserialize(sPath);

#endif

}

void CPS2OS::SaveExecutableConfig()
{

#ifdef DEBUGGER_INCLUDED

	char sPath[256];
	const char* sFilename;

	sFilename = Utils::GetFilenameFromPath(m_sExecutablePath);

	strcpy(sPath, CONFIGPATH);
	strcat(sPath, sFilename);
	strcat(sPath, ".functions");

	CPS2VM::m_EE.m_Functions.Serialize(sPath);

#endif

}

//////////////////////////////////////////////////
//Exception Handler
//////////////////////////////////////////////////

CPS2OS::CExceptionHandler::CExceptionHandler()
{
	m_nState		= STATE_START;
	m_pStateParam	= NULL;
}

CPS2OS::CExceptionHandler::~CExceptionHandler()
{

}

bool CPS2OS::CExceptionHandler::Execute()
{
	switch(m_nState)
	{
	case STATE_START:
		return Start();
		break;
	case STATE_DMACHANDLERS:
		return ProcessDmacHandlers();
		break;
	case STATE_DONE:
		return Done();
		break;
	}
	return false;
}

bool CPS2OS::CExceptionHandler::IsDone()
{
	return (m_nState == STATE_DONE);
}

bool CPS2OS::CExceptionHandler::Start()
{
	//Save the context
	memcpy(&m_Context, &CPS2VM::m_EE.m_State, sizeof(MIPSSTATE));

	//Disable interrupts
	CPS2VM::m_EE.m_State.nCOP0[CCOP_SCU::STATUS] &= ~0x0001;

	m_nState++;
	m_pStateParam = NULL;

	return true;
}

bool CPS2OS::CExceptionHandler::ProcessDmacHandlers()
{
	uint32 nINTCStatus;
	uint32 nDMACStatus;
	uint32 nDMACMask;
	bool bContinue;
	DMACHANDLER* pHandler;
	CList<DMACHANDLER>::ITERATOR* pIter;

	//Check some stuff first
	if(m_pStateParam == NULL)
	{
		bContinue = true;
		
		//Check if there's any handlers
		if(CPS2OS::m_DmacHandler.Count() == 0)
		{
			bContinue = false;
		}

		//Check if the exception handler was triggered by the DMAC
		nINTCStatus = CINTC::GetRegister(CINTC::INTC_STAT) & CINTC::GetRegister(CINTC::INTC_MASK);
		if(!(nINTCStatus & 0x0002))
		{
			bContinue = false;
		}

		if(!bContinue)
		{
			m_nState++;
			return true;
		}

		pIter = new CList<DMACHANDLER>::ITERATOR;
		(*pIter) = CPS2OS::m_DmacHandler.Begin();

		m_pStateParam = pIter;
	}
	else
	{
		pIter = (CList<DMACHANDLER>::ITERATOR*)m_pStateParam;
	}

	nDMACStatus	= CDMAC::GetRegister(CDMAC::D_STAT);
	nDMACMask	= (nDMACStatus >> 16);
	nDMACStatus &= nDMACMask;

	for(; !(*pIter); (*pIter)++)
	{
		pHandler = *(*pIter);
		if(((1 << pHandler->nChannel) & nDMACStatus) != 0)
		{
			//Jump to the interrupt handler
			ChangePC(pHandler->nAddress);
			CPS2OS::m_pCtx->m_State.nGPR0[31] = 0x1FC00004;

			(*pIter)++;
			return false;
		}
	}

	delete pIter;

	m_nState++;
	m_pStateParam = NULL;

	return true;
}

bool CPS2OS::CExceptionHandler::Done()
{
	//Restore context
	memcpy(&CPS2OS::m_pCtx->m_State, &m_Context, sizeof(MIPSSTATE));

	ChangePC(CPS2OS::m_pCtx->m_State.nPC);

	return false;
}

void CPS2OS::CExceptionHandler::ChangePC(uint32 nAddress)
{
	CPS2OS::m_pCtx->m_State.nPC = nAddress;
	CPS2OS::m_pCtx->m_nQuota = 1;
}

void CPS2OS::ExceptionHandler()
{
	CExceptionHandler* pHandler;

	//Check if interrupts are enabled
	if(!(m_pCtx->m_State.nCOP0[CCOP_SCU::STATUS] & 0x0001)) return;

	pHandler = new CExceptionHandler;
	m_ExceptionHandler.Insert(pHandler, m_ExceptionHandler.Count());

	ExceptionReentry();
}

void CPS2OS::ExceptionReentry()
{
	CExceptionHandler* pHandler;
	pHandler = m_ExceptionHandler.Find(m_ExceptionHandler.Count() - 1);

	if(pHandler == NULL)
	{
		//WTF?
		assert(0);
	}

	while(1)
	{
		if(!pHandler->Execute()) break;
	}

	if(pHandler->IsDone())
	{
		m_ExceptionHandler.Remove(m_ExceptionHandler.Count() - 1);
		delete pHandler;
	}
}

//////////////////////////////////////////////////
//System Calls
//////////////////////////////////////////////////

void CPS2OS::sc_Unhandled()
{
	printf("PS2OS: Unknown system call (%X) called from 0x%0.8X.\r\n", m_pCtx->m_State.nGPR0[3], m_pCtx->m_State.nPC);
}

//02
void CPS2OS::sc_GsSetCrt()
{
	CPS2VM::m_pGS->SetCrt(\
		m_pCtx->m_State.nGPR0[SC_PARAM0], \
		m_pCtx->m_State.nGPR0[SC_PARAM1], \
		m_pCtx->m_State.nGPR0[SC_PARAM2]);
}

//12
void CPS2OS::sc_AddDmacHandler()
{
	uint32 nAddress, nChannel, nNext, nID;
	DMACHANDLER* pHandler;

	nChannel	= m_pCtx->m_State.nGPR0[SC_PARAM0];
	nAddress	= m_pCtx->m_State.nGPR0[SC_PARAM1];
	nNext		= m_pCtx->m_State.nGPR0[SC_PARAM2];

	pHandler = new DMACHANDLER;
	pHandler->nChannel	= nChannel;
	pHandler->nAddress	= nAddress;

	m_nNextDmacHandlerID++;
	nID = m_nNextDmacHandlerID;

	m_DmacHandler.Insert(pHandler, nID);

	m_pCtx->m_State.nGPR0[SC_RETURN] = nID;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//16
void CPS2OS::sc_EnableDmac()
{
	uint32 nChannel, nRegister;

	nChannel = m_pCtx->m_State.nGPR0[SC_PARAM0];

	nRegister = (1 << nChannel) | (1 << (nChannel + 16));

	CDMAC::SetRegister(CDMAC::D_STAT, nRegister);

	//Enable INT1
	CINTC::SetRegister(0x1000F010, 0x00000002);

	m_pCtx->m_State.nGPR0[SC_RETURN] = 0;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//3C
void CPS2OS::sc_RFU060()
{
	uint32 nStackBase, nStackSize, nStackAddr;

	nStackBase = m_pCtx->m_State.nGPR0[SC_PARAM1];
	nStackSize = m_pCtx->m_State.nGPR0[SC_PARAM2];

	if(nStackBase == 0xFFFFFFFF)
	{
		nStackAddr = 0x02000000;
	}
	else
	{
		nStackAddr = nStackBase + nStackSize;
	}

	m_nStackBase = nStackAddr - nStackSize;

	m_pCtx->m_State.nGPR0[SC_RETURN] = nStackAddr;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//3D
void CPS2OS::sc_RFU061()
{
	uint32 nHeapBase, nHeapSize;

	nHeapBase = m_pCtx->m_State.nGPR0[SC_PARAM0];
	nHeapSize = m_pCtx->m_State.nGPR0[SC_PARAM1];

	if(nHeapSize == 0xFFFFFFFF)
	{
		m_nHeapBase = m_nStackBase;
	}
	else
	{
		m_nHeapBase = nHeapBase + nHeapSize;
	}

	m_pCtx->m_State.nGPR0[SC_RETURN] = m_nHeapBase;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//3E
void CPS2OS::sc_EndOfHeap()
{
	m_pCtx->m_State.nGPR0[SC_RETURN] = m_nHeapBase;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//40
void CPS2OS::sc_CreateSema()
{
	SEMAPHORE* pSrcSema;
	SEMAPHORE* pSema;
	uint32 nSemaID;

	pSrcSema = (SEMAPHORE*)(CPS2VM::m_pRAM + m_pCtx->m_State.nGPR0[SC_PARAM0]);

	pSema = new SEMAPHORE;
	memcpy(pSema, pSrcSema, sizeof(SEMAPHORE));

	m_nNextSemaphoreID++;
	nSemaID = m_nNextSemaphoreID;

	m_Semaphore.Insert(pSema, nSemaID);

	//Initialize the semaphore
	pSema->nCount = pSema->nInitCount;

	m_pCtx->m_State.nGPR0[SC_RETURN] = nSemaID;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//41
void CPS2OS::sc_DeleteSema()
{
	uint32 nSemaID;
	SEMAPHORE* pSema;

	nSemaID = m_pCtx->m_State.nGPR0[SC_PARAM0];

	pSema = m_Semaphore.Find(nSemaID);
	if(pSema == NULL)
	{
		assert(0);
	}

	//Check if any threads are waiting for this?
	m_Semaphore.Remove(nSemaID);
	delete pSema;

	//??
	m_pCtx->m_State.nGPR0[SC_RETURN] = 0;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//42
//43
void CPS2OS::sc_SignalSema()
{
	//Nothing to do for now really...
}

//44
void CPS2OS::sc_WaitSema()
{
	//Nothing to do for now really...
}

//64
void CPS2OS::sc_FlushCache()
{

}

//71
void CPS2OS::sc_GsPutIMR()
{
	
}

//76
void CPS2OS::sc_SifDmaStat()
{
	m_pCtx->m_State.nGPR0[SC_RETURN] = 0xFFFFFFFF;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0xFFFFFFFF;
}

//77
void CPS2OS::sc_SifSetDma()
{
	struct DMAREG
	{
		uint32 nSrcAddr;
		uint32 nDstAddr;
		uint32 nSize;
		uint32 nFlags;
	};

	DMAREG* pXfer;
	uint32 nCount, i, nSize;

	pXfer	= (DMAREG*)(CPS2VM::m_pRAM + m_pCtx->m_State.nGPR0[SC_PARAM0]);
	nCount	= m_pCtx->m_State.nGPR0[SC_PARAM1];

	//Returns count
	//DMA might call an interrupt handler
	m_pCtx->m_State.nGPR0[SC_RETURN] = nCount;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;

	for(i = 0; i < nCount; i++)
	{
		nSize = (pXfer[i].nSize + 0x0F) / 0x10;

		CDMAC::SetRegister(CDMAC::D6_MADR,	pXfer[i].nSrcAddr);
		CDMAC::SetRegister(CDMAC::D6_TADR,	pXfer[i].nDstAddr);
		CDMAC::SetRegister(CDMAC::D6_QWC,	nSize);
		CDMAC::SetRegister(CDMAC::D6_CHCR,	0x00000100);
	}
}

//78
void CPS2OS::sc_SifSetDChain()
{
	//Humm, set the SIF0 DMA channel in destination chain mode?
}

//79
void CPS2OS::sc_SifSetReg()
{
	uint32 nRegister, nValue;

	nRegister	= m_pCtx->m_State.nGPR0[SC_PARAM0];
	nValue		= m_pCtx->m_State.nGPR0[SC_PARAM1];

	CSIF::SetRegister(nRegister, nValue);

	m_pCtx->m_State.nGPR0[SC_RETURN] = 0;
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//7A
void CPS2OS::sc_SifGetReg()
{
	uint32 nRegister;

	nRegister = m_pCtx->m_State.nGPR0[SC_PARAM0];

	m_pCtx->m_State.nGPR0[SC_RETURN] = CSIF::GetRegister(nRegister);
	m_pCtx->m_State.nGPR1[SC_RETURN] = 0;
}

//////////////////////////////////////////////////
//System Call Handler
//////////////////////////////////////////////////

void CPS2OS::SysCallHandler()
{
	uint8 nFunc;

	nFunc = (uint8)(m_pCtx->m_State.nGPR0[3] & 0xFF);
	if(nFunc & 0x80)
	{
		nFunc = 0 - nFunc;
	}

	m_pSysCall[nFunc]();

	DisassembleSysCall(nFunc);
}

void CPS2OS::DisassembleSysCall(uint8 nFunc)
{
	if(!CPS2VM::m_Logging.GetOSLoggingStatus()) return;

	switch(nFunc)
	{
	case 0x02:
		printf("PS2OS: GsSetCrt(interlace = %i, mode = %i, field = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1], \
			m_pCtx->m_State.nGPR0[SC_PARAM2]);
		break;
	case 0x12:
		printf("PS2OS: AddDmacHandler(channel = %i, address = 0x%0.8X, next = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1], \
			m_pCtx->m_State.nGPR0[SC_PARAM2]);
		break;
	case 0x16:
		printf("PS2OS: EnableDmac(channel = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x3C:
		printf("PS2OS: RFU060(gp = 0x%0.8X, stack = 0x%0.8X, stack_size = 0x%0.8X, args = 0x%0.8X, root_func = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1], \
			m_pCtx->m_State.nGPR0[SC_PARAM2], \
			m_pCtx->m_State.nGPR0[SC_PARAM3], \
			m_pCtx->m_State.nGPR0[SC_PARAM4]);
		break;
	case 0x3D:
		printf("PS2OS: RFU061(heap_start = 0x%0.8X, heap_size = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1]);
		break;
	case 0x3E:
		printf("PS2OS: EndOfHeap();\r\n");
		break;
	case 0x40:
		printf("PS2OS: CreateSema(sema = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x41:
		printf("PS2OS: DeleteSema(semaid = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x43:
		printf("PS2OS: iSignalSema(semaid = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x44:
		printf("PS2OS: WaitSema(semaid = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x64:
#ifdef _DEBUG
//		printf("FlushCache();\r\n");
#endif
		break;
	case 0x71:
		printf("PS2OS: GsPutIMR(GS_IMR = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	case 0x76:
		printf("PS2OS: SifDmaStat();\r\n");
		break;
	case 0x77:
		printf("PS2OS: SifSetDma(list = 0x%0.8X, count = %i);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1]);
		break;
	case 0x78:
		printf("PS2OS: SifSetDChain();\r\n");
		break;
	case 0x79:
		printf("PS2OS: SifSetReg(register = 0x%0.8X, value = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0], \
			m_pCtx->m_State.nGPR0[SC_PARAM1]);
		break;
	case 0x7A:
		printf("PS2OS: SifGetReg(register = 0x%0.8X);\r\n", \
			m_pCtx->m_State.nGPR0[SC_PARAM0]);
		break;
	}
}

//////////////////////////////////////////////////
//System Call Handlers Table
//////////////////////////////////////////////////

void (*CPS2OS::m_pSysCall[0x80])() = 
{
	//0x00
	sc_Unhandled,		sc_Unhandled,		sc_GsSetCrt,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x08
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x10
	sc_Unhandled,		sc_Unhandled,		sc_AddDmacHandler,	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_EnableDmac,		sc_Unhandled,
	//0x18
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x20
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x28
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x30
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x38
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_RFU060,			sc_RFU061,			sc_EndOfHeap,		sc_Unhandled,
	//0x40
	sc_CreateSema,		sc_DeleteSema,		sc_SignalSema,		sc_SignalSema,		sc_WaitSema,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x48
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x50
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x58
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x60
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_FlushCache,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x68
	sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
	//0x70
	sc_Unhandled,		sc_GsPutIMR,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_SifDmaStat,		sc_SifSetDma,
	//0x78
	sc_SifSetDChain,	sc_SifSetReg,		sc_SifGetReg,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,		sc_Unhandled,
};
