#ifndef _PS2OS_H_
#define _PS2OS_H_

#include "ELF.h"
#include "Str.h"
#include "Slot.h"
#include "List.h"
#include "MIPS.h"

class CPS2OS
{
public:
												CPS2OS();
												~CPS2OS();

	static void									Initialize();
	static void									Release();
	static bool									IsInitialized();

	static void									BootELF(const char*);
	static CELF*								GetELF();

	static void									ExceptionHandler();
	static void									ExceptionReentry();

	static Framework::CSlot<int>				m_OnExecutableChange;
	static Framework::CSlot<int>				m_OnExecutableUnloading;

private:
	struct DMACHANDLER
	{
		uint32									nAddress;
		uint32									nChannel;
	};

	class CExceptionHandler
	{
	public:
												CExceptionHandler();
												~CExceptionHandler();
		bool									Execute();
		bool									IsDone();

	private:
		enum STATE
		{
			STATE_START = 0,
			STATE_DMACHANDLERS,
			STATE_DONE,
		};
		
		bool									Start();
		bool									ProcessDmacHandlers();
		bool									Done();

		void									ChangePC(uint32);

		MIPSSTATE								m_Context;
		unsigned int							m_nState;
		void*									m_pStateParam;
	};

	struct SEMAPHORE
	{
		uint32									nCount;
		uint32									nMaxCount;
		uint32									nInitCount;
		uint32									nWaitThreads;
		uint32									nAttributes;
		uint32									nOptions;
	};

	enum SYSCALL_REGS
	{
		SC_RETURN = 2,
		SC_PARAM0 = 4,
		SC_PARAM1 = 5,
		SC_PARAM2 = 6,
		SC_PARAM3 = 7,
		SC_PARAM4 = 8
	};

	static void									LoadExecutable();
	static void									UnloadExecutable();

	static void									SaveExecutableConfig();
	static void									LoadExecutableConfig();

	static void									SysCallHandler();
	static void									DisassembleSysCall(uint8);

	static void									(*m_pSysCall[0x80])();

	//Various system calls
	static void									sc_GsSetCrt();
	static void									sc_AddDmacHandler();
	static void									sc_EnableDmac();
	static void									sc_RFU060();
	static void									sc_RFU061();
	static void									sc_EndOfHeap();
	static void									sc_CreateSema();
	static void									sc_DeleteSema();
	static void									sc_SignalSema();
	static void									sc_WaitSema();
	static void									sc_FlushCache();
	static void									sc_GsPutIMR();
	static void									sc_SifDmaStat();
	static void									sc_SifSetDma();
	static void									sc_SifSetDChain();
	static void									sc_SifSetReg();
	static void									sc_SifGetReg();
	static void									sc_Unhandled();

	static Framework::CList<SEMAPHORE>			m_Semaphore;
	static uint32								m_nNextSemaphoreID;

	static Framework::CList<DMACHANDLER>		m_DmacHandler;
	static uint32								m_nNextDmacHandlerID;

	static Framework::CList<CExceptionHandler>	m_ExceptionHandler;

	//This should be stored into a thread's (?) context
	static uint32								m_nStackBase;
	static uint32								m_nHeapBase;

	static bool									m_nInitialized;

	static CELF*								m_pELF;
	static CMIPS*								m_pCtx;
	static Framework::CStr						m_sExecutablePath;
};

#endif
