#ifndef _PS2VM_H_
#define _PS2VM_H_

#include "Types.h"
#include "PS2OS.h"
#include "pthread.h"
#include "MIPS.h"
#include "ThreadMsg.h"
#include "GSHandler.h"
#include "PadHandler.h"
#include "LogControl.h"
#include "Slot.h"
#include "IOP_PadMan.h"

enum PS2VM_MSG
{
	PS2VM_MSG_NULL,
	PS2VM_MSG_PAUSE,
	PS2VM_MSG_RESUME,
	PS2VM_MSG_DESTROY,
	PS2VM_MSG_CREATEGS,
	PS2VM_MSG_DESTROYGS,
	PS2VM_MSG_CREATEPAD,
	PS2VM_MSG_DESTROYPAD,
	PS2VM_MSG_SAVESTATE,
	PS2VM_MSG_LOADSTATE,
};

enum PS2VM_STATUS
{
	PS2VM_STATUS_PAUSED,
	PS2VM_STATUS_RUNNING,
};

#define VERSION_MAJOR		(0)
#define VERSION_MINOR		(10)

typedef CGSHandler*			(*GSHANDLERFACTORY)(void*); 
typedef CPadHandler*		(*PADHANDLERFACTORY)(IOP::CPadMan*, void*);

class CPS2VM
{
public:
	enum PS2VM_RAMSIZE
	{
		RAMSIZE = 0x02000000,
	};

	enum PS2VM_BIOSSIZE
	{
		BIOSSIZE = 0x00400000,
	};

	static void						Initialize();
	static void						Destroy();

	static void						Resume();
	static void						Pause();

	static void						CreateGSHandler(GSHANDLERFACTORY, void*);
	static void						DestroyGSHandler();

	static void						CreatePadHandler(PADHANDLERFACTORY, void*);
	static void						DestroyPadHandler();

	static void						SaveState(const char*);
	static void						LoadState(const char*);

	static uint32					IOPortReadHandler(uint32);
	static void						IOPortWriteHandler(uint32, uint32);

	static CGSHandler*				m_pGS;
	static CPadHandler*				m_pPad;

	static uint8*					m_pRAM;
	static uint8*					m_pBIOS;

	static CMIPS					m_EE;
	static PS2VM_STATUS				m_nStatus;
	static unsigned int				m_nVBlankTicks;
	static bool						m_nInVBlank;

	static unsigned int				EETickFunction(unsigned int);

	static Framework::CSlot<int>	m_OnMachineStateChange;
	static Framework::CSlot<int>	m_OnRunningStateChange;
	static Framework::CSlot<int>	m_OnNewFrame;

	static CLogControl				m_Logging;

private:
	struct CREATEGSHANDLERPARAM
	{
		GSHANDLERFACTORY			pFactory;
		void*						pParam;
	};

	struct CREATEPADHANDLERPARAM
	{
		PADHANDLERFACTORY			pFactory;
		void*						pParam;
	};

	static void						CreateVM();
	static void						DestroyVM();
	static void						SaveVMState(const char*);
	static void						LoadVMState(const char*);

	static void						LoadBIOS();

	static void						SendMessage(PS2VM_MSG, void* = NULL);
	static void*					EmuThread(void*);
	static pthread_t				m_nThread;
	static CThreadMsg				m_MsgBox;

	static CPS2OS*					m_pOS;
};

#endif
