#include <stdio.h>
#include <string.h>
#include "RegViewFPU.h"
#include "PS2VM.h"

#define CLSNAME		_X("CRegViewFPU")
#define XMARGIN		6

using namespace Framework;

CRegViewFPU::CRegViewFPU(HWND hParent, RECT* pR, CMIPS* pC)
{
	if(!DoesWindowClassExist(CLSNAME))
	{
		WNDCLASSEX w;
		memset(&w, 0, sizeof(WNDCLASSEX));
		w.cbSize		= sizeof(WNDCLASSEX);
		w.lpfnWndProc	= CWindow::WndProc;
		w.lpszClassName	= CLSNAME;
		w.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
		w.hInstance		= GetModuleHandle(NULL);
		w.hCursor		= LoadCursor(NULL, IDC_ARROW);
		RegisterClassEx(&w);
	}

	m_nViewMode = VIEWMODE_SINGLE;

	Create(WS_EX_CLIENTEDGE, CLSNAME, _X(""), WS_VISIBLE | WS_CHILD, pR, hParent, NULL);
	SetClassPtr();

	m_pCtx = pC;
	
	m_pOnMachineStateChangeHandler = new CSlotHandlerMethod<CRegViewFPU, int>(this, &CRegViewFPU::OnMachineStateChange);
	m_pOnRunningStateChangeHandler = new CSlotHandlerMethod<CRegViewFPU, int>(this, &CRegViewFPU::OnRunningStateChange);

	CPS2VM::m_OnMachineStateChange.InsertHandler(m_pOnMachineStateChangeHandler);
	CPS2VM::m_OnRunningStateChange.InsertHandler(m_pOnRunningStateChangeHandler);
}

CRegViewFPU::~CRegViewFPU()
{
	CPS2VM::m_OnMachineStateChange.RemoveHandler(m_pOnMachineStateChangeHandler);
	CPS2VM::m_OnRunningStateChange.RemoveHandler(m_pOnRunningStateChangeHandler);
}

void CRegViewFPU::OnRunningStateChange(int nNothing)
{
	Redraw();
}

void CRegViewFPU::OnMachineStateChange(int nNothing)
{
	Redraw();
}

HFONT CRegViewFPU::GetFont()
{
	return CreateFont(-11, 0, 0, 0, 400, 0, 0, 0, 0, 1, 2, 1, 49, _X("Courier New"));
}

void CRegViewFPU::RenderFCSR(HDC hDC, unsigned int* pY)
{
	xchar sText[256];

	xsprintf(sText, _X("FCSR: 0x%0.8X"), m_pCtx->m_State.nFCSR);

	TextOut(hDC, XMARGIN, (*pY), sText, (int)xstrlen(sText));

	(*pY) += 18;

	xsprintf(sText, _X("CC  : %i%i%i%i%i%i%i%ib"), \
		(m_pCtx->m_State.nFCSR & 0x80000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x40000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x20000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x10000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x04000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x08000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x02000000) != 0 ? 1 : 0, \
		(m_pCtx->m_State.nFCSR & 0x00800000) != 0 ? 1 : 0);

	TextOut(hDC, XMARGIN, (*pY), sText, (int)xstrlen(sText));

	(*pY) += 18;
}

void CRegViewFPU::RenderWord(HDC hDC)
{
	char sText[256];
	char sReg[256];
	xchar sConvert[256];

	MIPSSTATE* s;
	int i;
	unsigned int nX, nY;
	uint32 nData;

	s = &m_pCtx->m_State;
	nY = 3;

	for(i = 0; i < 32; i++)
	{
		if(i < 10)
		{
			sprintf(sReg, "F%i  ", i);
		}
		else
		{
			sprintf(sReg, "F%i ", i);
		}

		nData = ((uint32*)s->nCOP10)[i * 2];

		sprintf(sText, "%s: 0x%0.8X", sReg, nData);

		nX = XMARGIN;

		xconvert(sConvert, sText, 256);

		TextOut(hDC, nX, nY, sConvert, (int)xstrlen(sConvert));

		nY += 18;
	}

	RenderFCSR(hDC, &nY);

}

void CRegViewFPU::RenderSingle(HDC hDC)
{
	char sText[256];
	char sReg[256];
	xchar sConvert[256];

	MIPSSTATE* s;
	int i;
	unsigned int nX, nY;
	float nValue;
	uint32 nData;

	s = &m_pCtx->m_State;
	nY = 3;

	for(i = 0; i < 32; i++)
	{
		if(i < 10)
		{
			sprintf(sReg, "F%i  ", i);
		}
		else
		{
			sprintf(sReg, "F%i ", i);
		}

		nData = ((uint32*)s->nCOP10)[i * 2];
		nValue = *(float*)(&nData);

		sprintf(sText, "%s: %+.24e", sReg, nValue);

		nX = XMARGIN;

		xconvert(sConvert, sText, 256);

		TextOut(hDC, nX, nY, sConvert, (int)xstrlen(sConvert));

		nY += 18;
	}

	RenderFCSR(hDC, &nY);

}

long CRegViewFPU::OnRightButtonUp(unsigned int nX, unsigned int nY)
{
	POINT pt;
	HMENU hMenu;

	pt.x = nX;
	pt.y = nY;
	ClientToScreen(m_hWnd, &pt);

	hMenu = CreatePopupMenu();
	InsertMenu(hMenu, 0, MF_BYPOSITION | (m_nViewMode == 0 ? MF_CHECKED : 0), 40000 + 0, _X("32 Bits Integers"));
	InsertMenu(hMenu, 1, MF_BYPOSITION | (m_nViewMode == 1 ? MF_CHECKED : 0), 40000 + 1, _X("64 Bits Integers"));
	InsertMenu(hMenu, 2, MF_BYPOSITION | (m_nViewMode == 2 ? MF_CHECKED : 0), 40000 + 2, _X("Single Precision Floating-Point Numbers"));
	InsertMenu(hMenu, 3, MF_BYPOSITION | (m_nViewMode == 3 ? MF_CHECKED : 0), 40000 + 3, _X("Double Precision Floating-Point Numbers"));

	TrackPopupMenu(hMenu, 0, pt.x, pt.y, 0, m_hWnd, NULL); 

	return FALSE;
}

long CRegViewFPU::OnCommand(unsigned short nID, unsigned short nCmd, HWND hSender)
{
	if((nID >= 40000) && (nID < (40000 + VIEWMODE_MAX)))
	{
		m_nViewMode = (VIEWMODE)(nID - 40000);
		Redraw();
	}

	return TRUE;
}

void CRegViewFPU::Paint(HDC hDC)
{
	RECT rwin;
	HFONT nFont;

	GetClientRect(&rwin);

	BitBlt(hDC, 0, 0, rwin.right, rwin.bottom, NULL, 0, 0, WHITENESS);

	nFont = GetFont();

	SelectObject(hDC, nFont);

	switch(m_nViewMode)
	{
	case VIEWMODE_WORD:
		RenderWord(hDC);
		break;
	case VIEWMODE_SINGLE:
		RenderSingle(hDC);
		break;
	}

	DeleteObject(nFont);
}
