#include <stdio.h>
#include <string.h>
#include "Settings.h"
#include "xml/Writer.h"
#include "xml/Parser.h"
#include "xml/Utils.h"
#include "StdStream.h"

using namespace Framework;

CSettings::CSettings()
{
	CStdStream* pStream;
	Xml::CNode* pDocument;
	
	if(sizeof(SETTINGS) == 0)
	{
		return;
	}

	m_nInvalid = true;
	memset(&m_Settings, 0, sizeof(SETTINGS));

	try
	{
		pStream = new CStdStream(fopen("settings.xml", "rb"));
	}
	catch(...)
	{
		return;
	}

	pDocument = Xml::CParser::ParseDocument(pStream);
	if(pDocument == NULL)
	{
		return;
	}

#ifdef DEBUGGER_INCLUDED
	LoadDebuggerSettings(pDocument);
#endif

	m_nInvalid = false;

	delete pDocument;
	delete pStream;

}

CSettings::~CSettings()
{
	CStdStream* pStream;
	Xml::CNode* pDocument;
	Xml::CNode* pSettings;

	if(sizeof(SETTINGS) == 0)
	{
		return;
	}

	try
	{
		pStream = new CStdStream(fopen("settings.xml", "wb"));
	}
	catch(...)
	{
		return;
	}

	pSettings = new Xml::CNode("Settings", true);

#ifdef DEBUGGER_INCLUDED
	SaveDebuggerSettings(pSettings);
#endif

	pDocument = new Xml::CNode;
	pDocument->InsertNode(pSettings);

	Xml::CWriter::WriteDocument(pStream, pDocument);

	delete pDocument;
	delete pStream;
}

void CSettings::LoadDebuggerSettings(Xml::CNode* pDocument)
{

#ifdef DEBUGGER_INCLUDED

	//Log Window Settings
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/LogPosX",			&m_Settings.nLogPosX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/LogPosY",			&m_Settings.nLogPosY);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/LogSizeX",			&m_Settings.nLogSizeX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/LogSizeY",			&m_Settings.nLogSizeY);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/LogVisible",			&m_Settings.nLogVisible);

	//Disassembly Window Settings
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/DisAsmPosX",			&m_Settings.nDisAsmPosX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/DisAsmPosY",			&m_Settings.nDisAsmPosY);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/DisAsmSizeX",		&m_Settings.nDisAsmSizeX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/DisAsmSizeY",		&m_Settings.nDisAsmSizeY);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/DisAsmVisible",		&m_Settings.nDisAsmVisible);

	//Register View Window Settings
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/RegViewPosX",		&m_Settings.nRegViewPosX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/RegViewPosY",		&m_Settings.nRegViewPosY);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/RegViewSizeX",		&m_Settings.nRegViewSizeX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/RegViewSizeY",		&m_Settings.nRegViewSizeY);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/RegViewVisible",		&m_Settings.nRegViewVisible);

	//Memory View Window Settings
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/MemoryViewPosX",		&m_Settings.nMemoryViewPosX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/MemoryViewPosY",		&m_Settings.nMemoryViewPosY);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/MemoryViewSizeX",	&m_Settings.nMemoryViewSizeX);
	Xml::GetNodeIntValue(pDocument,		"Settings/Debugger/MemoryViewSizeY",	&m_Settings.nMemoryViewSizeY);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/MemoryViewVisible",	&m_Settings.nMemoryViewVisible);

	//Memory View Window Settings
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/GSLoggingEnabled",	&m_Settings.nGSLogging);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/DMACLoggingEnabled",	&m_Settings.nDMACLogging);
	Xml::GetNodeBoolValue(pDocument,	"Settings/Debugger/OSLoggingEnabled",	&m_Settings.nOSLogging);

#endif

}

void CSettings::SaveDebuggerSettings(Xml::CNode* pSettings)
{

#ifdef DEBUGGER_INCLUDED

	Xml::CNode* pDebugger;

	pDebugger = new Xml::CNode("Debugger", true);

	//Log Window Settings
	pDebugger->InsertNode(Xml::CreateNodeIntValue("LogPosX",				m_Settings.nLogPosX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("LogPosY",				m_Settings.nLogPosY));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("LogSizeX",				m_Settings.nLogSizeX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("LogSizeY",				m_Settings.nLogSizeY));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("LogVisible",			m_Settings.nLogVisible));

	//Disassembly Window Settings
	pDebugger->InsertNode(Xml::CreateNodeIntValue("DisAsmPosX",				m_Settings.nDisAsmPosX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("DisAsmPosY",				m_Settings.nDisAsmPosY));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("DisAsmSizeX",			m_Settings.nDisAsmSizeX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("DisAsmSizeY",			m_Settings.nDisAsmSizeY));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("DisAsmVisible",			m_Settings.nDisAsmVisible));

	//Register View Window Settings
	pDebugger->InsertNode(Xml::CreateNodeIntValue("RegViewPosX",			m_Settings.nRegViewPosX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("RegViewPosY",			m_Settings.nRegViewPosY));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("RegViewSizeX",			m_Settings.nRegViewSizeX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("RegViewSizeY",			m_Settings.nRegViewSizeY));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("RegViewVisible",		m_Settings.nRegViewVisible));

	//Memory View Window Settings
	pDebugger->InsertNode(Xml::CreateNodeIntValue("MemoryViewPosX",			m_Settings.nMemoryViewPosX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("MemoryViewPosY",			m_Settings.nMemoryViewPosY));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("MemoryViewSizeX",		m_Settings.nMemoryViewSizeX));
	pDebugger->InsertNode(Xml::CreateNodeIntValue("MemoryViewSizeY",		m_Settings.nMemoryViewSizeY));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("MemoryViewVisible",		m_Settings.nMemoryViewVisible));

	//Logging Settings
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("GSLoggingEnabled",		m_Settings.nGSLogging));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("DMACLoggingEnabled",	m_Settings.nDMACLogging));
	pDebugger->InsertNode(Xml::CreateNodeBoolValue("OSLoggingEnabled",		m_Settings.nOSLogging));
	
	pSettings->InsertNode(pDebugger);

#endif

}
